import traceback
from http import HTTPStatus
from signal import SIGKILL
from subprocess import TimeoutExpired
from time import time

# pylint: disable=import-error
from flask import jsonify, Response

# pylint: disable=no-name-in-module, import-error
import txtapi.utility.debugger.response_converter as converter
from txtapi.utility.debugger.debugger_commands import Command
from txtapi.utility.debugger.ft_debugger import FtDebugger
from txtapi.models.debugger_response import DebuggerResponse
from txtapi.models.error_message import ErrorMessage
from txtapi.models.expression import Expression
from txtapi.models.inline_response_default import InlineResponseDefault
from txtapi.models.program_location import ProgramLocation
from txtapi.utility.watch_process import WatchProcess
from txtapi.handlers import util_handler

# pylint: enable=import-error

debuggers = dict()

def add_breakpoints(workspace_name, breakpoints) -> (Response, int):
    
    try:  
        debugger = get_debugger(workspace_name)
        created_breakpoints = []

        for breakpoint_ in breakpoints: 
                    
            filename = breakpoint_.filename
            line = breakpoint_.line
            if filename is None or filename == "" or line is None or line == "":
                continue
                    
            try:
                command = "b " + filename + ":" + str(line)
                new_breakpoint = converter.convert_add_breakpoint_response(debugger.send_command(command))
                created_breakpoints.append(new_breakpoint)
            except Exception as ex:
                pass

        response = jsonify(created_breakpoints)
        status_code = HTTPStatus.OK

    except Exception as ex:
        response, status_code = create_500_response(str(ex), "add_breakpoints")   

    return response, status_code.value
    

def add_expressions(workspace_name, expressions):
    
    try:
        debugger = get_debugger(workspace_name)
        created_expressions = []

        for expression in expressions: 
                    
            key = expression.expression_key
            cmd = expression.command
            if key is None or key == "" or cmd is None or cmd == "":
                continue
                    
            command = Command.ALIAS.value + " _" + key + " " + cmd
            debuggerResponse = converter.remove_pdb_prompt(debugger.send_command(command))
            if debuggerResponse:
                continue
            
            add_type_to_expression(debugger, expression)
            created_expressions.append(expression)
            
        response = jsonify(created_expressions)
        status_code = HTTPStatus.OK

    except Exception as ex:
        response, status_code =  create_500_response(str(ex), "add_expressions")  
    
    return response, status_code.value


def call_continue(workspace_name) -> (Response, int):
    try:
        program_location = call_single_command(workspace_name, Command.CONT)
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "call_continue")
    return response, status_code.value


def callstack_down(workspace_name, steps):
    try:
        command = "{0} {1}".format(Command.DOWN.value, steps)
        program_location = call_jump_command(workspace_name, command)
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "callstack_down")
    return response, status_code.value


def callstack_up(workspace_name, steps):
    try:
        command = "{0} {1}".format(Command.UP.value, steps)
        program_location = call_jump_command(workspace_name, command)
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "callstack_up")
    return response, status_code.value


def delete_all_breakpoints(workspace_name) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)
        debugger.send_clear_command()

        response = jsonify(
            InlineResponseDefault(
                "Successfully deleted all breakpoints."
            )
        )
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "delete_all_breakpoints")
    
    return response, status_code.value


def delete_breakpoint_by_id(workspace_name, breakpoint_id) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)
        command = Command.CLEAR.value + " " + str(breakpoint_id)
        debugger.send_command(command)
  
        response = jsonify(
            InlineResponseDefault(
                "Successfully deleted breakpoint with ID " + str(breakpoint_id)
            )
        )
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "delete_breakpoint_by_id")
    
    return response, status_code.value


def delete_expression_by_key(workspace_name, expression_key):
    
    try:
        debugger = get_debugger(workspace_name)
        command = Command.UNALIAS.value + " _" + expression_key
            
        debuggerResponse = converter.remove_pdb_prompt(debugger.send_command(command))
        if debuggerResponse:
            raise Exception("Unexpected response from Debugger: " + debuggerResponse + "'")
            
        response = jsonify(
            InlineResponseDefault(
                "Successfully deleted expression with key = " + expression_key
            )
        )
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "delete_expression_by_key")

    return response, status_code.value


def get_breakpoints(workspace_name) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)
 
        debuggerResponse = debugger.send_command(Command.BREAK.value)
        breakpoints = converter.convert_get_all_breakpoints_response(debuggerResponse)  

        response = jsonify(breakpoints)
        status_code = HTTPStatus.OK
    
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "get_breakpoints")    
    
    return response, status_code.value


def get_callstack(workspace_name) -> (Response, int):
     
    try:
        debugger = get_debugger(workspace_name)

        debuggerResponse = debugger.send_command(Command.WHERE.value)
        callstack = converter.convert_callstack(debuggerResponse)
        
        response = jsonify(callstack)
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "get_callstack")
    
    return response, status_code.value


def get_expressions(workspace_name):
    
    try:
        debugger = get_debugger(workspace_name)
            
        debuggerResponse = debugger.send_command(Command.ALIAS.value)
        expressions = converter.convert_alias_list(debuggerResponse)
            
        for expression in expressions:
            add_type_to_expression(debugger, expression)
            
        response = jsonify(expressions)
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code  = create_500_response(str(ex), "get_expressions")
    
    return response, status_code.value


def get_globals(workspace_name):
    
    try:
        debugger = get_debugger(workspace_name)
            
        debuggerResponse = debugger.send_command(Command.GLOBALS.value)
        expressions = converter.convert_expressions(debuggerResponse)
            
        response = jsonify(expressions)
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "get_globals")
    
    return response, status_code.value


def get_locals(workspace_name):
    
    try:
        debugger = get_debugger(workspace_name)
            
        debuggerResponse = debugger.send_command(Command.LOCALS.value)
        expressions = converter.convert_expressions(debuggerResponse)
            
        response = jsonify(expressions)
        status_code = HTTPStatus.OK
        
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "get_locals")
    
    return response, status_code.value


def restart(workspace_name, debugger_arguments) -> (Response, int):
    return start(workspace_name, debugger_arguments)


def start(workspace_name, debugger_arguments) -> (Response, int):
    
    try:
        util_handler.stop()
    except:
        pass
    
    try:
        debugger = FtDebugger(workspace_name)
        debuggers[workspace_name] = debugger
        
        pdb_args = []
        if debugger_arguments is not None:
            if debugger_arguments.pdb_args is not None:
                pdb_args = debugger_arguments.pdb_args

        program_location, user_prints = converter.convert_debugger_response(debugger.start(pdb_args))
        WatchProcess.get_instance().send_clear()
        WatchProcess.get_instance().send_started()
        WatchProcess.get_instance().send_data(user_prints)

        if debugger_arguments is not None:
            if debugger_arguments.breakpoints is not None:
                add_breakpoints(workspace_name, debugger_arguments.breakpoints)
            if debugger_arguments.expressions is not None:
                add_expressions(workspace_name, debugger_arguments.expressions)

        response, status_code = create_debugger_response(workspace_name, program_location, True)

    except FileNotFoundError as ex:
        delete_debugger(workspace_name)
        response, status_code = create_404_response(str(ex), "start")
    except Exception as ex:
        delete_debugger(workspace_name)
        response, status_code = create_500_response(str(ex), "start")
    
    return response, status_code.value


def step_into(workspace_name) -> (Response, int):
    try:
        program_location = call_single_command(workspace_name, Command.STEP)
        if program_location is None:
            # the debugger is in a system methode (e.g. print)
            # he need twice 'return' and 'next' command to reach the next line
            for _ in range(2):
                call_single_command(workspace_name, Command.RETURN)
                program_location = call_single_command(workspace_name, Command.NEXT)
                if program_location is not None:
                    break
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "step_into")
    return response, status_code.value


def step_out(workspace_name) -> (Response, int):
    try:
        call_single_command(workspace_name, Command.RETURN)
        program_location = call_single_command(workspace_name, Command.NEXT)
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "step_out")
    return response, status_code.value


def step_over(workspace_name) -> (Response, int):
    try:
        program_location = call_single_command(workspace_name, Command.NEXT)
        response, status_code = create_debugger_response(workspace_name, program_location)
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "step_over")
    return response, status_code.value


def stop_debugger(workspace_name) -> (Response, int):
    try:
        debugger = get_debugger(workspace_name)
        returncode = end_debugger_process(debugger)
        response = InlineResponseDefault(
            "Successfully stopped debugger with workspace_name {0}. Process returncode: {1}".format(workspace_name, returncode)
        )
    except:
        response = InlineResponseDefault("No debugger with workspace_name " + workspace_name + " is running.")

    delete_debugger(workspace_name)
    
    return jsonify(response), HTTPStatus.OK.value


def stop_debuggers():
    for workspace_name in list(debuggers.keys()):
        stop_debugger(workspace_name)


def toggle_breakpoint(workspace_name, breakpoint_id, enabled) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)

        command = Command.ENABLE.value if enabled.enabled else Command.DISABLE.value
        command += " " + str(breakpoint_id)

        debuggerResponse = debugger.send_command(command)
        breakpoint_ = converter.convert_toggle_breakpoint_response(debuggerResponse, enabled.enabled)
        
        if not breakpoint_:
            response, status_code = create_404_response("No breakpoint with ID " + str(breakpoint_id) + " found.", "toggle_breakpoint")
        else:
            response = jsonify(breakpoint_)
            status_code = HTTPStatus.OK

    except Exception as ex:
        response, status_code = create_500_response(str(ex), "toggle_breakpoint")

    return response, status_code.value


def toggle_breakpoints(workspace_name, enabled) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)

        debuggerResponse = debugger.send_command(Command.BREAK.value)
        breakpoints = converter.convert_get_all_breakpoints_response(debuggerResponse)  

        if len(breakpoints) > 0:

            debugger = debuggers.get(workspace_name)
            
            command = Command.ENABLE.value if enabled.enabled else Command.DISABLE.value
            for breakpoint_ in breakpoints:
                command += " " + str(breakpoint_.id)

            debuggerResponse = debugger.send_command(command)
            breakpoints = converter.convert_toggle_multiple_bps(debuggerResponse, enabled.enabled)
        
        response = jsonify(breakpoints)
        status_code = HTTPStatus.OK
    
    except Exception as ex:
        response, status_code = create_500_response(str(ex), "toggle_breakpoints")
   
    return response, status_code.value


def update_expression(workspace_name, expression_key, command):
    return add_expressions(workspace_name, [Expression(expression_key=expression_key, command=command)])


# BEGIN OF DEBUGGER COMMAND METHODS

def call_single_command(workspace_name, command: Command) -> ProgramLocation:   
    
    debugger = get_debugger(workspace_name)
    debuggerResponse = debugger.send_command(command.value)
    program_location, user_prints = converter.convert_debugger_response(debuggerResponse)
    WatchProcess.get_instance().send_data(user_prints)

    return program_location


def call_jump_command(workspace_name, command: str) -> ProgramLocation:
    
    debugger = get_debugger(workspace_name)
    debuggerResponse = debugger.send_command(command)
    program_location, user_prints = converter.convert_jump_response(debuggerResponse)
    WatchProcess.get_instance().send_data(user_prints)
    
    return program_location


# BEGINN OF RESPONSE METHODS

def create_debugger_response(workspace_name, program_location: ProgramLocation, with_breakpoints = False) -> (Response, int):
    
    try:
        debugger = get_debugger(workspace_name)

        debuggerResponse = debugger.send_command(Command.ALIAS.value)
        expressions = converter.convert_alias_list(debuggerResponse)
        for expression in expressions:
            add_type_to_expression(debugger, expression)

        debuggerResponse = debugger.send_command(Command.WHERE.value)
        callstack = converter.convert_callstack(debuggerResponse)

        breakpoints = None
        if with_breakpoints:
            debuggerResponse = debugger.send_command(Command.BREAK.value)
            breakpoints = converter.convert_get_all_breakpoints_response(debuggerResponse)  

        response = jsonify(DebuggerResponse(
            program_location = program_location,
            callstack = callstack,
            expressions = expressions,
            breakpoints = breakpoints
        ))
        status_code = HTTPStatus.OK

    except Exception as ex:
        response, status_code = create_500_response(str(ex), "create_debugger_response")
    
    return response, status_code


def create_400_response(params: str, origin: str = "create_400_response") -> (Response, int):
    return jsonify(ErrorMessage(
        errorcode=HTTPStatus.BAD_REQUEST.phrase,
        message="Malformed request syntax: " + params,
        timestamp=time(),
        origin="debugger_handler." + origin
    )), HTTPStatus.BAD_REQUEST


def create_404_response(ex: str, origin: str = "create_404_response") -> (Response, int):
    return jsonify(ErrorMessage(
        errorcode=HTTPStatus.NOT_FOUND.phrase,
        message="Active debuggers: " + str(debuggers)
        + ". Exception: " + ex
        + ". " + traceback.format_exc(),
        timestamp=time(),
        origin="debugger_handler." + origin
    )), HTTPStatus.NOT_FOUND


def create_500_response(ex: str, origin: str = "create_500_response") -> (Response, int):
    return jsonify(ErrorMessage(
        errorcode=HTTPStatus.INTERNAL_SERVER_ERROR.phrase,
        message="Active debuggers: " + str(debuggers)
        + ". Exception: " + ex
        + ". " + traceback.format_exc(),
        timestamp=time(),
        origin="debugger_handler." + origin
    )), HTTPStatus.INTERNAL_SERVER_ERROR


# BEGINN OF DEBUGGER PROCESS METHODS


def delete_debugger(workspace_name: str):
    try:
        del debuggers[workspace_name]
    except Exception:
        pass


def end_debugger_process(debugger) -> int:
    try:
        debugger.send_command(Command.QUIT.value)
    except Exception:
        pass
    try:
        return debugger.process.wait(timeout=1)
    except Exception:
        debugger.process.kill()
        return SIGKILL


def get_debugger(workspace_name) -> FtDebugger:
    if workspace_name in debuggers and debuggers.get(workspace_name).is_socket_open:
        return debuggers.get(workspace_name)
    delete_debugger(workspace_name)
    raise Exception("No active debugging session with workspace_name " + workspace_name + " was found.")


# BEGINN OF HELPER METHODS

def add_type_to_expression(debugger, expression: Expression):
    """
    (Pdb) _akey  <- key
    123         <- value
    (Pdb) type(123)
    <class 'int'>   <- type
    (Pdb) error_example
    *** NameError: name 'error_example' is not defined
    """
    response = debugger.send_command("_" + expression.expression_key)
    expression.value = converter.remove_pdb_prompt(response)

    if expression.value.startswith("***"):
        expression.type = "Error"
    else:
        command = "type(" + str(expression.value) + ")"
        expression.type = converter.remove_pdb_prompt(debugger.send_command(command))

