import os
import ast
import json
import re
from typing import List

from txtapi.utility.debugger.rpdb_helper import EOT
from txtapi.environment.environment import WORKSPACE_BASEPATH
from txtapi.models.breakpoint import Breakpoint
from txtapi.models.expression import Expression
from txtapi.models.program_location import ProgramLocation

WORKSPACE_ABSPATH = os.path.abspath(WORKSPACE_BASEPATH)

def convert_debugger_response(response: str) -> (ProgramLocation, str):
    """
    Converts output from the debugger to the ProgramLocation model.
    Debugger output has the following format:
    [user print output]\n
    [--Return--|--Call--]\n
    > WORKSPACE_ABSPATH/workspace_name/filename(line)methodname()[->returntype]\n
    [->...\n]
    (Pdb) 

    Only the part between (the first occurence of) WORKSPACE_ABSPATH and '()\n' is relevant for ProgramLocation.
    The part before that is output printed by the debugged program.
    """

    if not response:
        return None, None
    pattern_program_location = r".+\(\d+\).+\(\)(->.+)?\n" # filename(line)methodname()[->returntype]\n
    pattern = r"(--(Return|Call)--\n)?> " + pattern_program_location + r"(.+\n)?" + re.escape(EOT) + r"\Z"

    user_prints = re.sub(pattern, "", response)
    
    match = re.search(pattern, response)
    if not match:
        return None, user_prints
    response = match.group(0)

    match = re.search(pattern_program_location, response)
    response = re.sub(r"\n\Z", "", match.group(0))
    program_location = extract_program_location(response)
    return program_location, user_prints


def convert_jump_response(response: str) -> ProgramLocation:
    if response.startswith("***"): # *** Newest frame / *** Oldest frame
        return None
    else:
        return convert_debugger_response(response)


def convert_callstack(response: str) -> List[ProgramLocation]:
    """
      /usr/lib/python3.7/bdb.py(585)run()
      <string>(1)<module>()
      /home/ftc/apps/workspace_name1/main.py(1)<module>()
    -> print("Hallo Welt")
    > /usr/lib/python3.7/codecs.py(328)reset()
    (Pdb) """
    lines = response.split("\n")
    pl_list = list()
    for line in lines:
        pl = extract_program_location(line)
        if pl:
            pl_list.append(pl)
    return pl_list


def convert_line_to_breakpoint(response: str) -> Breakpoint:
    """
    format:
    Num Type         Disp Enb   Where
    1   breakpoint   keep yes   at /home/ftc/apps/workspace_name1/main.py:1
    	breakpoint already hit 1 time
    2   breakpoint   keep yes   at /home/ftc/apps/workspace_name1/main.py:3
    6   breakpoint   keep yes   at /home/ftc/apps/workspace_name1/main.py:6
    (Pdb)
    """
    pattern_entire_line = r"\d+ +breakpoint +\w+ +(yes|no) +at +" + re.escape(WORKSPACE_ABSPATH) + r".+:\d+"
    match = re.search(pattern_entire_line, response)
    if not match:
        return None
    response = match.group(0)

    pattern = re.escape(WORKSPACE_ABSPATH)
    parts = re.split(pattern, response, maxsplit=1)

    filename, line = extract_filename_line(parts[1])

    match = re.search(r"yes|no", parts[0])
    enabled = True if match.group(0) == "yes" else False

    match = re.search(r"\A\d+", response) # \A matches start of string
    i_d = match.group(0)

    bp = Breakpoint(id=int(i_d), filename=filename, line=int(line), enabled=enabled)
    return bp


def convert_add_breakpoint_response(response: str) -> Breakpoint:
    """
    Successfully setting the breakpoint gives response with the following format:
        Breakpoint i_d at WORKSPACE_ABSPATH/workspace_name/filename.py:line
    Errors can cause responses like these:
        *** Blank or comment
        *** 'main.py' not found from sys.path
        End of file
    """
    response = remove_pdb_prompt(response)
    pattern_entire_line = r"\ABreakpoint \d+ at " + re.escape(WORKSPACE_ABSPATH) + r".+:\d+"
    match = re.search(pattern_entire_line, response)
    if not match:
        return None
    response = match.group(0)

    match = re.search(r"\d+", response)
    if not match:
        return None
    i_d = match.group(0)

    pattern = r"(?<=" + re.escape(WORKSPACE_ABSPATH) + r").+" # matches everything after WORKSPACE_ABSPATH
    match = re.search(pattern, response)
    if not match:
        return None
    filename_and_line = match.group(0)

    filename, line = extract_filename_line(filename_and_line)
    bp = Breakpoint(id=int(i_d), filename=filename, line=int(line), enabled=True)
    return bp


def convert_toggle_breakpoint_response(response: str, enabled: bool) -> Breakpoint:
    """
    format:
    Disabled breakpoint i_d at WORKSPACE_ABSPATH/workspace_name/filename.py:line
    Enabled breakpoint i_d at WORKSPACE_ABSPATH/workspace_name/filename.py:line
    *** Breakpoint 2 already deleted
    *** Breakpoint number 9 out of range
    *** Non-numeric breakpoint number xyz
    """

    pattern_entire_line = r"(Dis|En)abled breakpoint \d+ at " + re.escape(WORKSPACE_ABSPATH) + r".*:\d+"
    match = re.search(pattern_entire_line, response)
    if not match:
        return None
    # get only the matching string without additional lines or line breaks
    response = match.group(0)

    match = re.search(r"(?<=abled breakpoint )\d+", response)
    if not match:
        return None
    i_d = match.group(0)
    # without "(Dis|En)" because look-behind requires fixed-width pattern

    match = re.search(r"(?<=" + re.escape(WORKSPACE_ABSPATH) + ").+", response)
    if not match:
        return None
    filename_and_line = match.group(0)

    filename, line = extract_filename_line(filename_and_line)
    bp = Breakpoint(id=int(i_d), filename=filename, line=int(line), enabled=enabled)
    return bp


def convert_expressions(response: str) -> List[Expression]:
    response = remove_pdb_prompt(response)
    response_dict = ast.literal_eval(response)
    expressions_list = list()
    for key, value in response_dict.items():
        value = json.loads(value)
        expression = Expression(expression_key=json.loads(key), value=value[0], type=value[1])
        expressions_list.append(expression)
    return expressions_list


def extract_program_location(response_line: str) -> ProgramLocation:
    """
    WORKSPACE_ABSPATH/workspace_name/filename.py(line)methodname()[->returntype]
    <frozen importlib._bootstrap>(980)_find_and_load()
    /usr/lib/python3.7/codecs.py(327)reset()
    """
    current_frame = False
    if response_line.startswith(">"):
        current_frame = True

    pattern_line_end = r"\(\)(->.+)?\Z"
    pattern_entire_line = r".+\(\d+\).+" + pattern_line_end # file(line)methodname()[->returntype]
    match = re.search(pattern_entire_line, response_line)
    if not match:
        return None
    
    response_line = re.sub(pattern_line_end, "", match.group(0))
    methodname = response_line.rsplit(")", maxsplit=1)[1]
    
    filename_and_line = re.sub(re.escape(methodname) + r"\Z", "", response_line)
    filename_and_line = filename_and_line.rsplit("(", maxsplit=1)
    filename = filename_and_line[0]
    line = re.sub(r"\)\Z", "", filename_and_line[1])

    if WORKSPACE_ABSPATH in filename:
        filename = filename.split(WORKSPACE_ABSPATH, maxsplit=1)[1]
        filename = base_filename(filename)
    else:
        return None

    return ProgramLocation(methodname=methodname, filename=filename, line=int(line), current_frame=bool(current_frame))


def convert_alias_list(response: str) -> List[Expression]:
    """
    format:
    _key = command
    """
    response = remove_pdb_prompt(response)
    pattern = " = "
    lines = response.split("\n")
    expressions = list()
    for line in lines:
        if pattern in line:
            key, command = re.split(pattern, line, maxsplit=1)
            key = key[1:] # cut off the _ prefix
            e = Expression(expression_key=key, command=command)
            expressions.append(e)
    return expressions


def convert_toggle_multiple_bps(response: str, enabled: bool) -> List[Breakpoint]:
    breakpoints_list = list()
    lines = response.split("\n")
    for line in lines:
        bp = convert_toggle_breakpoint_response(line, enabled)
        if bp is not None:
            breakpoints_list.append(bp)
    return breakpoints_list


def convert_get_all_breakpoints_response(response: str) -> List[Breakpoint]:
    breakpoints_list = list()
    lines = response.split("\n")
    for line in lines:
        bp = convert_line_to_breakpoint(line)
        if bp is not None:
            breakpoints_list.append(bp)
    return breakpoints_list


def extract_filename_line(filename_and_line: str) -> (str, int):
    parts = filename_and_line.rsplit(":", maxsplit=1)
    line = parts[1]
    filename = base_filename(parts[0])
    return filename, line


def remove_pdb_prompt(response: str) -> str:
    pattern = r"\n?" + re.escape(EOT) + r"\Z"
    # \n?   matches 0 or 1 newline
    # \Z    matches the end of the string
    response = re.sub(pattern, "", response)
    return response


def base_filename(path: str):
    # removes slashes at the beginning and at the end of the string
    path = re.sub("(^\/|\/$)", "", path.strip())
    parts = path.split("/")
    parts.pop(0)
    filename = "/".join(parts)
    return filename
    